import torch


def check_tensor(tensor, dtype):
    """
    Convert :code:`tensor` into a :code:`dtype` torch.Tensor.

    :param numpy.array/torch.tensor tensor: Input data.
    :param str dtype: PyTorch dtype string.
    :return: A :code:`dtype` torch.Tensor.
    """
    return torch.tensor(tensor, dtype=dtype)


def reset_params(model):
    """
    Reset all parameters in :code:`model`.
    :param torch.nn.Module model: Pytorch model.
    """
    if hasattr(model, "reset_parameters"):
        model.reset_parameters()
    else:
        for layer in model.children():
            reset_params(layer)


class NumpyDataLoader:
    """
    Convert numpy arrays into a dataloader.

    :param numpy.array *inputs: Numpy arrays.
    """
    def __init__(self, *inputs):
        self.inputs = inputs
        self.n_inputs = len(inputs)

    def __len__(self):
        return self.inputs[0].shape[0]

    def __getitem__(self, item):
        if self.n_inputs == 1:
            return self.inputs[0][item]
        else:
            return [array[item] for array in self.inputs]


import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from scipy.special import softmax
from torch.utils.data import DataLoader

from .utils import NumpyDataLoader


def ur_loss(frs, tau=0.5):
    """

    The uniform regularization (UR) proposed by Cui et al. [3].
    UR loss is computed as :math:`\ell_{UR} = \sum_{r=1}^R (\frac{1}{N}\sum_{n=1}^N f_{n,r} - \tau)^2`,
    where :math:`f_{n,r}` represents the firing level of the :math:`n`-th sample on the :math:`r`-th rule.

    :param torch.tensor frs: The firing levels (output of the antecedent) with the size of :math:`[N, R]`,
    where :math:`N` is the number of samples, :math:`R` is the number of ruels.
    :param float tau: The expectation :math:`\tau` of the average firing level for each rule. For a
        :math:`C`-class classification problem, we recommend setting :math:`\tau` to :math:`1/C`,
        for a regression problem, :math:`\tau` can be set as :math:`0.5`.
    :return: A scale value, representing the UR loss.


    """
    return ((torch.mean(frs, dim=0) - tau) ** 2).sum()


class Wrapper:
    """

    This class provide a training framework for beginners to train their fuzzy neural networks.

    :param torch.nn.Module model: The pre-defined TSK model.
    :param torch.Optimizer optimizer: Pytorch optimizer.
    :param torch.nn._Loss: Pytorch loss. For example, :code:`torch.nn.CrossEntropyLoss()` for classification tasks,
        and :code:`torch.nn.MSELoss()` for regression tasks.
    :param int batch_size: Batch size during training & prediction.
    :param int epochs: Training epochs.
    :param [Callback] callbacks: List of callbacks.
    :param str label_type: Label type, "c" or "r", when :code:`label_type="c"`, label's dtype will be changed to
        "int64", when :code:`label_type="r"`, label's dtype will be changed to "float32".

    Examples
    --------
    >>> from pytsk.gradient_descent import antecedent_init_center, AntecedentGMF, TSK, EarlyStoppingACC, EvaluateAcc, Wrapper
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.metrics import accuracy_score
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.preprocessing import StandardScaler
    >>> from torch.optim import AdamW
    >>> import torch.nn as nn
    >>> # ----------------- define data -----------------
    >>> X, y = make_classification(random_state=0)
    >>> x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.3)
    >>> ss = StandardScaler()
    >>> x_train = ss.fit_transform(x_train)
    >>> x_test = ss.transform(x_test)
    >>> # ----------------- define TSK model -----------------
    >>> n_rule = 10  # define number of rules
    >>> n_class = 2  # define output dimension
    >>> order = 1  # first-order TSK is used
    >>> consbn = True  # consbn tech is used
    >>> weight_decay = 1e-8  # weight decay for pytorch optimizer
    >>> lr = 0.01  # learning rate for pytorch optimizer
    >>> init_center = antecedent_init_center(x_train, y_train, n_rule=n_rule)  # obtain the initial antecedent center
    >>> gmf = AntecedentGMF(in_dim=x_train.shape[1], n_rule=n_rule, high_dim=True, init_center=init_center)  # define antecedent
    >>>  model = TSK(in_dim=x_train.shape[1], out_dim=n_class, n_rule=n_rule, antecedent=gmf, order=order, consbn=consbn)  # define TSK
    >>> # ----------------- define optimizers -----------------
    >>> ante_param, other_param = [], []
    >>> for n, p in model.named_parameters():
    >>>     if "center" in n or "sigma" in n:
    >>>         ante_param.append(p)
    >>>     else:
    >>>         other_param.append(p)
    >>> optimizer = AdamW(
    >>>     [{'params': ante_param, "weight_decay": 0},  # antecedent parameters usually don't need weight_decay
    >>>     {'params': other_param, "weight_decay": weight_decay},],
    >>>     lr=lr
    >>> )
    >>> # ----------------- split 20% data for earlystopping -----------------
    >>> x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.2)
    >>> # ----------------- define the earlystopping callback -----------------
    >>> EACC = EarlyStoppingACC(x_val, y_val, verbose=1, patience=40, save_path="tmp.pkl")  # Earlystopping
    >>> TACC = EvaluateAcc(x_test, y_test, verbose=1)  # Check test acc during training
    >>> # ----------------- train model -----------------
    >>> wrapper = Wrapper(model, optimizer=optimizer, criterion=nn.CrossEntropyLoss(),
    >>>               epochs=300, callbacks=[EACC, TACC], ur=0, ur_tau=1/n_class)  # define training wrapper, ur weight is set to 0
    >>> wrapper.fit(x_train, y_train)  # fit
    >>> wrapper.load("tmp.pkl")  # load best model saved by EarlyStoppingACC callback
    >>> y_pred = wrapper.predict(x_test).argmax(axis=1)  # predict, argmax for extracting classfication label
    >>> print("[TSK] ACC: {:.4f}".format(accuracy_score(y_test, y_pred)))  # print ACC
    """
    def __init__(self, model, optimizer, criterion,
                 batch_size=512, epochs=1, callbacks=None, label_type="c",
                 device="cpu", reset_param=True, ur=0, ur_tau=0.5, **kwargs):
        self.model = model
        self.optimizer = optimizer
        self.criterion = criterion
        self.batch_size = batch_size
        self.epochs = epochs
        self.device = device
        self.model.to(self.device)
        self.label_type = label_type

        self.ur = ur
        self.ur_tau = ur_tau

        if callbacks is None:
            self.callbacks = []
        elif isinstance(callbacks, list):
            self.callbacks = callbacks
        else:
            raise ValueError("callback must be a Callback object")
        self.reset_param = reset_param
        if self.reset_param:
            self.model.reset_parameters()
        self.cur_batch = 0
        self.cur_epoch = 0

        self.kwargs = kwargs

    def train_on_batch(self, input, target):
        """

        Define how to update a model with one batch of data.
        This method can be overwrite for custom training strategy.

        :param torch.tensor input: Feature matrix with the size of :math:`[N,D]`,
            :math:`N` is the number of samples, :math:`D` is the input dimension.
        :param torch.tensor target: Target matrix with the size of :math:`[N,C]`,
            :math:`C` is the output dimension.

        """
        # update model once
        input, target = input.to(self.device), target.to(self.device)
        outputs, frs = self.model(input, get_frs=True)
        if self.ur > 0:
            ur_loss_value = ur_loss(frs, self.ur_tau)
            loss = self.criterion(outputs, target) + self.ur * ur_loss_value
        else:
            loss = self.criterion(outputs, target)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    def fit(self, X, y):
        """

        Train the :code:`model` with numpy array.

        :param numpy.array X: Feature matrix :math:`X` with the size of :math:`[N, D]`.
        :param numpy.array y: Label matrix :math:`Y` with the size of :math:`[N, C]`,
            for classification task, :math:`C=1`, for regression task, :math:`C` is the
            number of the output dimension of :code:`model`.

        """
        X = X.astype("float32")
        if self.label_type == "c":
            y = y.astype("int64")
        elif self.label_type == "r":
            y = y.astype("float32")
        else:
            raise ValueError("label_type can only be \"c\" or \"r\"!")

        train_loader = DataLoader(
            NumpyDataLoader(X, y),
            batch_size=self.batch_size,
            shuffle=self.kwargs.get("shuffle", True),
            num_workers=self.kwargs.get("num_workers", 0),
            drop_last=self.kwargs.get("drop_last", True if self.batch_size < X.shape[0] else False),
        )

        self.fit_loader(train_loader)
        return self

    def fit_loader(self, train_loader):
        """

        Train the :code:`model` with user-defined pytorch dataloader.

        :param torch.utils.data.DataLoader train_loader: Data loader, the
            output of the loader should be corresponding to the inputs of :func:`train_on_batch <train_on_batch>`.
            For example, if dataloader has two output, then :func:`train_on_batch <train_on_batch>`
            should also have two inputs.

        """
        self.stop_training = False
        for e in range(self.epochs):
            self.cur_epoch = e
            self.__run_callbacks__("on_epoch_begin")
            for inputs in train_loader:
                self.__run_callbacks__("on_batch_begin")
                self.model.train()
                self.train_on_batch(inputs[0], inputs[1])
                self.__run_callbacks__("on_batch_end")
                self.cur_batch += 1
            self.__run_callbacks__("on_epoch_end")
            if self.stop_training:
                break
        return self

    def predict(self, X, y=None):
        """

        Get the prediction of the model.

        :param numpy.array X: Feature matrix :math:`X` with the size of :math:`[N, D]`.
        :param y: Not used.
        :return: Prediction matrix :math:`\hat{Y}` with the size of :math:`[N, C]`,
            :math:`C` is the output dimension of the :code:`model`.

        """
        X = X.astype("float32")
        test_loader = DataLoader(
            NumpyDataLoader(X),
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.kwargs.get("num_workers", 0),
            drop_last=False
        )
        y_preds = []
        for inputs in test_loader:
            self.model.eval()
            y_pred = self.model(inputs.to(self.device)).detach().cpu().numpy()
            y_preds.append(y_pred)
        return np.concatenate(y_preds, axis=0)

    def predict_proba(self, X, y=None):
        """

        For classification problem only, need :code:`label_type="c"`, return the prediction after softmax.

        :param numpy.array X: Feature matrix :math:`X` with the size of :math:`[N, D]`.
        :param y: Not used.
        :return: Prediction matrix :math:`\hat{Y}` with the size of :math:`[N, C]`,
            :math:`C` is the output dimension of the :code:`model`.

        """
        if self.label_type == "r":
            raise ValueError("predict_proba can only be used when label_type=\"c\"")
        y_preds = self.predict(X)
        return softmax(y_preds, axis=1)

    def __run_callbacks__(self, func_name):
        for cb in self.callbacks:
            getattr(cb, func_name)(self)

    def save(self, path):
        """

        Save model.

        :param str path: Model save path.
        """
        torch.save(self.model.state_dict(), path)

    def load(self, path):
        """

        Load model.

        :param str path: Model save path.
        """
        self.model.load_state_dict(torch.load(path))



from .antecedent import antecedent_init_center, Antecedent, AntecedentShareGMF, AntecedentGMF
from .callbacks import Callback, EarlyStoppingACC, EvaluateAcc
from .tsk import TSK
from .training import Wrapper, ur_loss
from .utils import NumpyDataLoader

__all__ = [
    "antecedent_init_center",
    "Antecedent",
    "AntecedentGMF",
    "AntecedentShareGMF",
    "Callback",
    "EvaluateAcc",
    "EarlyStoppingACC",
    "ur_loss",
    "Wrapper",
    "TSK",
    "NumpyDataLoader"
]


import importlib.util

package_name = "torch"
spec = importlib.util.find_spec(package_name)
if spec is None:
    print("Warning! PyTorch package is not installed, "
          "this package must be installed when you need to use pytsk.gradient_descent")



import torch
import torch.nn as nn

from .utils import reset_params


class TSK(nn.Module):
    """

    Parent: :code:`torch.nn.Module`

    This module define the consequent part of the TSK model and combines it with a pre-defined
     antecedent module. The input of this module is the raw feature matrix, and output
     the final prediction of a TSK model.

    :param int in_dim: Number of features :math:`D`.
    :param int out_dim: Number of output dimension :math:`C`.
    :param int n_rule: Number of rules :math:`R`, must equal to the :code:`n_rule` of
        the :code:`Antecedent()`.
    :param torch.Module antecedent: An antecedent module, whose output dimension should be
        equal to the number of rules :math:`R`.
    :param int order: 0 or 1. The order of TSK. If 0, zero-oder TSK, else, first-order TSK.
    :param float eps: A constant to avoid the division zero error.
    :param torch.nn.Module consbn: If none, the raw feature will be used as the consequent input;
        If a pytorch module, then the consequent input will be the output of the given module.
        If you wish to use the BN technique we mentioned in
        `Models & Technique <../models.html#batch-normalization>`_,  you can set
        :code:`precons=nn.BatchNorm1d(in_dim)`.
    """
    def __init__(self, in_dim, out_dim, n_rule, antecedent, order=1, eps=1e-8, precons=None):
        super(TSK, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.n_rule = n_rule
        self.antecedent = antecedent
        self.precons = precons

        self.order = order
        assert self.order == 0 or self.order == 1, "Order can only be 0 or 1."
        self.eps = eps

        self.__build_model__()

    def __build_model__(self):
        if self.order == 0:
            self.cons = nn.Linear(self.n_rule, self.out_dim, bias=True)
        else:
            self.cons = nn.Linear((self.in_dim + 1) * self.n_rule, self.out_dim)

    def reset_parameters(self):
        """
        Re-initialize all parameters, including both consequent and antecedent parts.

        :return:
        """
        reset_params(self.antecedent)
        self.cons.reset_parameters()

        if self.precons is not None:
            self.precons.reset_parameters()

    def forward(self, X, get_frs=False):
        """

        :param torch.tensor X: Input matrix with the size of :math:`[N, D]`,
            where :math:`N` is the number of samples.
        :param bool get_frs: If true, the firing levels (the output of the antecedent)
            will also be returned.

        :return: If :code:`get_frs=True`, return the TSK output :math:`Y\in \mathbb{R}^{N,C}`
            and the antecedent output :math:`U\in \mathbb{R}^{N,R}`. If :code:`get_frs=False`,
            only return the TSK output :math:`Y`.
        """
        frs = self.antecedent(X)

        if self.precons is not None:
            X = self.precons(X)

        if self.order == 0:
            cons_input = frs
        else:
            X = X.unsqueeze(dim=1).expand([X.size(0), self.n_rule, X.size(1)])  # [n_batch, n_rule, in_dim]
            X = X * frs.unsqueeze(dim=2)
            X = X.view([X.size(0), -1])
            cons_input = torch.cat([X, frs], dim=1)

        output = self.cons(cons_input)
        if get_frs:
            return output, frs
        return output






from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="scikit-pytsk",
    version="0.0.6",
    packages=find_packages(),
    url="https://github.com/YuqiCui/PyTSK",
    license="MIT",
    author="Yuqi Cui",
    author_email="yqcui@qq.com",
    description="PyTSK provides tools for uses conveniently developing TSK fuzzy systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["numpy", "scipy", "scikit-learn"],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)


import numpy as np
from scipy.spatial.distance import cdist
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


class BaseFuzzyClustering(object):
    def set_params(self, **params):
        """
        Setting attributes. Implemented to adapt the API of scikit-learn.
        """
        for p, v in params.items():
            setattr(self, p, v)
        return self


def __get_fuzzy_index__(m, N, D):
    if m == 'auto' and min(N, D - 1) >= 3:
        m_ = min(N, D - 1) / (min(N, D - 1) - 2)
    elif isinstance(m, float) or isinstance(m, int):
        m_ = float(m)
    else:
        m_ = 2
        print("Warning: auto set does not satisfied min(N, D - 1) >= 3, "
              "min(N, D - 1) = {}, setting m to 2".format(min(N, D - 1)))
    return m_


def __normalize_column__(X):
    X = np.fmax(X, np.finfo(np.float64).eps)
    return X / np.sum(X, axis=0, keepdims=True)


def __fcm_update__(data, u_old, m, dist="euclidean"):
    u_old = __normalize_column__(u_old)
    u_old = np.fmax(u_old, np.finfo(np.float64).eps)
    um = u_old ** m

    cntr = np.dot(um, data) / um.sum(axis=1, keepdims=True)
    d = cdist(data, cntr, metric=dist).T
    d = np.fmax(d, np.finfo(np.float64).eps)

    loss = np.sum(um * d ** 2)
    u = __normalize_column__(d ** (2 / (1 - m)))
    return cntr, u, loss, d


def __fcm_predict__(data, cntr, m, dist="euclidean"):
    d = cdist(data, cntr, metric=dist).T
    d = np.fmax(d, np.finfo(np.float64).eps)
    u = __normalize_column__(d ** (2 / (1 - m)))
    return u


def x2xp(X, U, order):
    """
    Convert matrix :math:`X` and :math:`U` into the consequent input matrix :math:`X_p`

    :param numpy.array x: size: :math:`[N,D]`. Input features.
    :param numpy.array u: size: :math:`[N,R]`. Corresponding membership degree matrix.
    :param int order: 0 or 1. The order of TSK models.
    :return: If :code:`order=0`, return :math:`U` directly, elif :code:`order=1`,
        return the matrix :math:`X_p` with the size of :math:`[N, (D+1)\times R]`.
        Details can be found at [2].
    [2] Wang S, Chung K F L, Zhaohong D, et al. Robust fuzzy clustering neural network
        based on ɛ-insensitive loss function[J]. Applied Soft Computing, 2007, 7(2): 577-584.
    """
    assert order == 0 or order == 1, "Order can only be 0 or 1."
    R = U.shape[1]
    if order == 1:
        N = X.shape[0]
        mem = np.expand_dims(U, axis=1)
        X = np.expand_dims(np.concatenate((X, np.ones([N, 1])), axis=1), axis=2)
        X = np.repeat(X, repeats=R, axis=2)
        xp = X * mem
        xp = xp.reshape([N, -1])
        return xp
    else:
        return U


def compute_variance(X, U, V):
    """

    Convert the feature matrix :math:`X` and the membership degree matrix :math:`U`
    into the consequent input matrix :math:`X_p`

    Each row in :math:`X\in \mathbb{R}^{N,D}` represents a :math:`D`-dimension input
    vector. Suppose vector :math:`\mathbf{x}` is one row, and then the consequent input
     matrix :math:`P` is computed as [5] for a first-order TSK:

    .. math::
        &\mathbf{x}_e = (1, \mathbf{x}),\\
        &\tilde{\mathbf{x}}_r = u_r \mathbf{x}_e,\\
        &\mathbf{p} = (\tilde{\mathbf{x}}_1, \tilde{\mathbf{x}}_2, ...,\tilde{\mathbf{x}}_R),

    where :math:`\mathbf{p}` is the corresponding row in :math:`P`, which is a
    :math:`(D+1)\times R`-dimension vector. Then the consequent parameters of TSK can
    be optimized by any linear regression algorithms.


    :param numpy.array x: size: :math:`[N,D]`. Input features.
    :param numpy.array u: size: :math:`[N,R]`. Corresponding membership degree matrix.
    :param int order: 0 or 1. The order of TSK models.
    :return: If :code:`order=0`, return :math:`U` directly, else if :code:`order=1`,
        return the matrix :math:`X_p` with the size of :math:`[N, (D+1)\times R]`.
        Details can be found at [2].

    [2] `Wang S, Chung K F L, Zhaohong D, et al. Robust fuzzy clustering neural network based
        on ɛ-insensitive loss function[J]. Applied Soft Computing, 2007, 7(2): 577-584.
        <https://www.sciencedirect.com/science/article/pii/S1568494606000469>`_
    """
    R = U.shape[0]
    D = X.shape[1]
    variance = np.zeros([R, D])
    for i in range(D):
        variance[:, i] = np.sum(
            U * ((X[:, i][:, np.newaxis] - V[:, i].T) ** 2).T, axis=1
        ) / np.sum(U, axis=1)
    return variance


class FuzzyCMeans(BaseFuzzyClustering, BaseEstimator, TransformerMixin):
    """
    The fuzzy c-means (FCM) clustering algorithm [1]. This implementation is adopted
    from the `scikit-fuzzy <https://pythonhosted.org/scikit-fuzzy/overview.html>`_ package.
    When constructing a TSK fuzzy system, a fuzzy clustering algorithm is usually used to
    compute the antecedent parameters, after that, the consequent parameters can be computed
    by least-squared error algorithms, such as Ridge regression [2]. How to use this class
     can be found at `Quick start <quick_start.html#training-with-fuzzy-clustering>`_.

    The objective function of the FCM is:

    .. math::
        &J = \sum_{i=1}^{N}\sum_{j=1}^{C} U_{i,j}^m\|\mathbf{x}_i - \mathbf{v}_j\|_2^2\\
        &s.t. \sum_{j=1}^{C}\mu_{i,j} = 1, i = 1,...,N,

    :param int n_cluster: Number of clusters, equal to the number of rules :math:`R` of a TSK model.
    :param float/str fuzzy_index: Fuzzy index of the FCM algorithm, default `auto`. If
        :code:`fuzzy_index=auto`, then the fuzzy index is computed as :math:`\min(N, D-1) / (\min(N, D-1)-2)`
        (If :math:`\min(N, D-1)<3`, fuzzy index will be set to 2), according to [3]. Otherwise the given
        float value is used.
    :param float/str sigma_scale: The scale parameter :math:`h` to adjust the actual standard deviation
        :math:`\sigma` of the Gaussian membership function in TSK antecedent part. If :code:`sigma_scale=auto`,
        :code:`sigma_scale` will be set as :math:`\sqrt{D}`, where :math:`D` is the input dimension [4].
        Otherwise the given float value is used.
    :param str/np.array init: The initialization strategy of the membership grid matrix :math:`U`.
        Support "random" or numpy array with the size of :math:`[R, N]`, where :math:`R` is the number of
        clusters/rules, :math:`N` is the number of training samples. If :code:`init="random"`, the initial
        membership grid matrix will be randomly initialized, otherwise the given matrix will be used.
    :param int tol_iter: The total iteration of the FCM algorithm.
    :param float error: The maximum error that will stop the iteration before maximum iteration is reached.
    :param str dist: The distance type for the :func:`scipy.spatial.distance.cdist` function, default
        "euclidean". The distance function can also be "braycurtis", "canberra", "chebyshev", "cityblock",
        "correlation", "cosine", "dice", "euclidean", "hamming", "jaccard", "jensenshannon", "kulsinski",
        "kulczynski1", "mahalanobis", "matching", "minkowski", "rogerstanimoto", "russellrao", "seuclidean",
        "sokalmichener", "sokalsneath", "sqeuclidean", "yule".
    :param int verbose: If > 0, it will show the loss of the FCM objective function during iterations.
    :param int order: 0 or 1. Decide whether to construct a zero-order TSK or a first-order TSK.
    """

    def __init__(self, n_cluster, fuzzy_index="auto", sigma_scale="auto",
                 init="random", tol_iter=100, error=1e-6, dist="euclidean",
                 verbose=0, order=1):
        self.n_cluster = n_cluster
        self.fuzzy_index = fuzzy_index
        self.sigma_scale = sigma_scale
        self.init = init
        self.tol_iter = tol_iter
        assert self.tol_iter > 1, "tol_iter must > 1."
        self.error = error
        self.dist = dist
        self.verbose = verbose
        self.order = order

    def get_params(self, deep=True):
        return {
            "n_cluster": self.n_cluster,
            "fuzzy_index": self.fuzzy_index,
            "sigma_scale": self.sigma_scale,
            "init": self.init,
            "tol_iter": self.tol_iter,
            "error": self.error,
            "dist": self.dist,
            "verbose": self.verbose,
            "order": self.order,
        }

    def fit(self, X, y=None):
        """
        Run the FCM algorithm.

        :param numpy.array X: Input array with the size of :math:`[N, D]`, where :math:`N` is the number
            of training samples, and :math:`D` is number of features.
        :param numpy.array y: Not used. Pass None.
        """
        check_array(X, ensure_2d=True)
        N, D = X.shape[0], X.shape[1]

        self.m_ = __get_fuzzy_index__(self.fuzzy_index, N, D)
        self.scale_ = np.sqrt(D) if self.sigma_scale == "auto" else self.sigma_scale
        self.n_features = D

        if self.init == "random":
            u = np.random.rand(self.n_cluster, N)
            u = __normalize_column__(u)
        elif isinstance(self.init, np.ndarray):
            u = __normalize_column__(self.init)
        else:
            raise ValueError("Unsupported init param, must be \"random\" or "
                             "numpy.ndarray of size [n_clusters, n_samples]")
        self.iter_cnt = 0
        cntr = None
        self.loss_hist = []
        for t in range(self.tol_iter):
            uold = u.copy()
            cntr, u, loss, d = __fcm_update__(X, uold, self.m_, self.dist)
            change = np.linalg.norm(uold - u, ord=2)
            self.loss_hist.append(loss)
            if self.verbose > 0:
                print('[FCM Iter {}] Loss: {:.4f}, change: {:.4f}'.format(t, loss, change))
            if change < self.error:
                break
            self.iter_cnt += 1
        self.fitted = True
        self.cluster_centers_ = cntr
        self.membership_degrees_ = u
        self.variance_ = compute_variance(X, self.membership_degrees_, self.cluster_centers_) * self.scale_

    def predict(self, X, y=None):
        """
        Predict the membership degrees of :code:`X` on each cluster.

        :param numpy.array X: Input array with the size of :math:`[N, D]`, where :math:`N` is the number
            of training samples, and :math:`D` is number of features.
        :param numpy.array y: Not used. Pass None.
        :return: return the membership degree matrix :math:`U` with the size of :math:`[N, R]`, where :math:`N`
            is the number of samples of :code:`X`, and :math:`R` is the number of clusters/rules. :math:`U_{i,j}`
            represents the membership degree of the :math:`i`-th sample on the :math:`r`-th cluster.
        """
        u = __fcm_predict__(X, self.cluster_centers_, self.m_, self.dist)
        return u.T

    def fit_transform(self, X, y=None, **fit_params):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        """
        Compute the membership degree matrix :math:`U`, and use :math:`X` and :math:`U` to get the consequent
        input matrix :math:`P` using function :func:`x2xp(x, u, order) <x2xp>`

        :param numpy.array X: Input array with the size of :math:`[N, D]`, where :math:`N` is the number of
            training samples, and :math:`D` is number of features.
        :param numpy.array y: Not used. Pass None.
        :return: return the consequent input :math:`P` with the size of :math:`[N, (D+1)\times R]`, where
            :math:`N` is the number of test samples, :math:`D` is number of features, :math:`R` is the number
            of clusters/rules.
        """
        d = -(np.expand_dims(X, axis=2) - np.expand_dims(self.cluster_centers_.T, axis=0)) ** 2 \
            / (2 * self.variance_.T + 1e-12)
        d = np.exp(np.sum(d, axis=1))
        d = np.fmax(d, 1e-16)
        u = d / np.sum(d, axis=1, keepdims=True)
        return x2xp(X, u, self.order)



from sklearn.metrics import accuracy_score


class Callback:
    """
    Similar as the callback class in Keras, our package provides a simplified version
    of callback, which allow users to monitor metrics during the training.
    We strongly recommend uses to custom their callbacks, here we provide two
    examples, :func:`EvaluateAcc <EvaluateAcc>` and :func:`EarlyStoppingACC <EarlyStoppingACC>`.
    """
    def on_batch_begin(self, wrapper):
        pass

    def on_batch_end(self, wrapper):
        pass

    def on_epoch_begin(self, wrapper):
        pass

    def on_epoch_end(self, wrapper):
        pass


class EvaluateAcc(Callback):
    """

    Evaluate the accuracy during training.

    :param numpy.array X: Feature matrix with the size of :math:`[N, D]`.
    :param numpy.array y: Label matrix with the size of :math:`[N, 1]`.
    """
    def __init__(self, X, y, verbose=0):
        super(EvaluateAcc, self).__init__()
        self.X, self.y = X, y
        self.verbose = verbose
        self.logs = []

    def on_epoch_end(self, wrapper):
        cur_log = {}
        y_pred = wrapper.predict(self.X).argmax(axis=1)
        acc = accuracy_score(y_true=self.y, y_pred=y_pred)
        cur_log["epoch"] = wrapper.cur_epoch
        cur_log["acc"] = acc
        self.logs.append(cur_log)
        if self.verbose > 0:
            print("[Epoch {:5d}] Test ACC: {:.4f}".format(cur_log["epoch"], cur_log["acc"]))


class EarlyStoppingACC(Callback):
    """

    Early-stopping by classification accuracy.

    :param numpy.array X: Feature matrix with the size of :math:`[N, D]`.
    :param numpy.array y: Label matrix with the size of :math:`[N, 1]`.
    :param int patience: Number of epochs with no improvement after which training will be stopped.
    :param int verbose: verbosity mode.
    :param str save_path: If :code:`save_path=None`, do not save models, else save the model
        with the best accuracy to the given path.
    """
    def __init__(self, X, y, patience=1, verbose=0, save_path=None):
        super(EarlyStoppingACC, self).__init__()
        self.X, self.y = X, y
        self.verbose = verbose
        self.patience = patience
        self.best_acc = 0
        self.cnt = 0
        self.logs = []
        self.save_path = save_path

    def on_epoch_end(self, wrapper):
        cur_log = {}
        y_pred = wrapper.predict(self.X).argmax(axis=1)
        acc = accuracy_score(y_true=self.y, y_pred=y_pred)

        if acc > self.best_acc:
            self.best_acc = acc
            self.cnt = 0
            if self.save_path is not None:
                wrapper.save(self.save_path)
        else:
            self.cnt += 1
            if self.cnt > self.patience:
                wrapper.stop_training = True
        cur_log["epoch"] = wrapper.cur_epoch
        cur_log["acc"] = acc
        cur_log["best_acc"] = self.best_acc
        self.logs.append(cur_log)
        if self.verbose > 0:
            print("[Epoch {:5d}] EarlyStopping Callback ACC: {:.4f}, Best ACC: {:.4f}".format(cur_log["epoch"], cur_log["acc"], cur_log["best_acc"]))




import itertools
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.init as init
from sklearn.cluster import KMeans
from torch.nn import functional

from .utils import check_tensor
import importlib.util

# For illustrative purposes.
package_name = 'pandas'

spec = importlib.util.find_spec(package_name)
if spec is None:
    print(package_name +" is not installed")


def antecedent_init_center(X, y=None, n_rule=2, method="kmean", engine="sklearn", n_init=20):
    """

    This function run KMeans clustering to obtain the :code:`init_center` for :func:`AntecedentGMF() <AntecedentGMF>`.

    Examples
    --------
    >>> init_center = antecedent_init_center(X, n_rule=10, method="kmean", n_init=20)
    >>> antecedent = AntecedentGMF(X.shape[1], n_rule=10, init_center=init_center)


    :param numpy.array X: Feature matrix with the size of :math:`[N,D]`, where :math:`N` is the
        number of samples, :math:`D` is the number of features.
    :param numpy.array y: None, not used.
    :param int n_rule: Number of rules :math:`R`. This function will run a KMeans clustering to
        obtain :math:`R` cluster centers as the initial antecedent center for TSK modeling.
    :param str method: Current version only support "kmean".
    :param str engine: "sklearn" or "faiss". If "sklearn", then the :code:`sklearn.cluster.KMeans()`
        function will be used, otherwise the :code:`faiss.Kmeans()` will be used. Faiss provide a
        faster KMeans clustering algorithm, "faiss" is recommended for large datasets.
    :param int n_init: Number of initialization of the KMeans algorithm. Same as the parameter
        :code:`n_init` in :code:`sklearn.cluster.KMeans()` and the parameter :code:`nredo` in
        :code:`faiss.Kmeans()`.
    """
    def faiss_cluster_center(X, y=None, n_rule=2, n_init=20):
        import faiss
        km = faiss.Kmeans(d=X.shape[1], k=n_rule, nredo=n_init)
        km.train(np.ascontiguousarray(X.astype("float32")))
        centers = km.centroids.T
        return centers

    if method == "kmean":
        if engine == "faiss":
            package_name = "faiss"
            spec = importlib.util.find_spec(package_name)
            if spec is not None:
                center = faiss_cluster_center(X=X, y=y, n_rule=n_rule)
                return center
            else:
                print("Package " + package_name + " is not installed, running scikit-learn KMeans...")
        km = KMeans(n_clusters=n_rule, n_init=n_init)
        km.fit(X)
        return km.cluster_centers_.T


class Antecedent(nn.Module):
    def forward(self, **kwargs):
        raise NotImplementedError

    def init(self, **kwargs):
        raise NotImplementedError

    def reset_parameters(self):
        raise NotImplementedError


class AntecedentGMF(Antecedent):
    """

    Parent: :code:`torch.nn.Module`

    The antecedent part with Gaussian membership function. Input: data, output the corresponding
    firing levels of each rule. The firing level :math:`f_r(\mathbf{x})` of the
    :math:`r`-th rule are computed by:

    .. math::
        &\mu_{r,d}(x_d) = \exp(-\frac{(x_d - m_{r,d})^2}{2\sigma_{r,d}^2}),\\
        &f_{r}(\mathbf{x})=\prod_{d=1}^{D}\mu_{r,d}(x_d),\\
        &\overline{f}_r(\mathbf{x}) = \frac{f_{r}(\mathbf{x})}{\sum_{i=1}^R f_{i}(\mathbf{x})}.


    :param int in_dim: Number of features :math:`D` of the input.
    :param int n_rule: Number of rules :math:`R` of the TSK model.
    :param bool high_dim: Whether to use the HTSK defuzzification. If :code:`high_dim=True`,
        HTSK is used. Otherwise the original defuzzification is used. More details can be found at [1].
        TSK model tends to fail on high-dimensional problems, so set :code:`high_dim=True` is highly
         recommended for any-dimensional problems.
    :param numpy.array init_center: Initial center of the Gaussian membership function with
        the size of :math:`[D,R]`. A common way is to run a KMeans clustering and set
        :code:`init_center` as the obtained centers. You can simply run
        :func:`pytsk.gradient_descent.antecedent.antecedent_init_center <antecedent_init_center>`
        to obtain the center.
    :param float init_sigma: Initial :math:`\sigma` of the Gaussian membership function.
    :param float eps: A constant to avoid the division zero error.
    """
    def __init__(self, in_dim, n_rule, high_dim=False, init_center=None, init_sigma=1., eps=1e-8):
        super(AntecedentGMF, self).__init__()
        self.in_dim = in_dim
        self.n_rule = n_rule
        self.high_dim = high_dim

        self.init_center = check_tensor(init_center, torch.float32) if init_center is not None else None
        self.init_sigma = init_sigma
        self.zr_op = torch.mean if high_dim else torch.sum
        self.eps = eps

        self.__build_model__()

    def __build_model__(self):
        self.center = nn.Parameter(torch.zeros(size=(self.in_dim, self.n_rule)))
        self.sigma = nn.Parameter(torch.zeros(size=(self.in_dim, self.n_rule)))

        self.reset_parameters()

    def init(self, center, sigma):
        """

        Change the value of :code:`init_center` and :code:`init_sigma`.

        :param numpy.array center: Initial center of the Gaussian membership function with the
            size of :math:`[D,R]`. A common way is to run a KMeans clustering and set
            :code:`init_center` as the obtained centers. You can simply run
            :func:`pytsk.gradient_descent.antecedent.antecedent_init_center <antecedent_init_center>`
            to obtain the center.
        :param float sigma: Initial :math:`\sigma` of the Gaussian membership function.
        """
        center = check_tensor(center, torch.float32)
        self.init_center = center
        self.init_sigma = sigma

        self.reset_parameters()

    def reset_parameters(self):
        """
        Re-initialize all parameters.

        :return:
        """
        init.constant_(self.sigma, self.init_sigma)

        if self.init_center is not None:
            self.center.data[...] = torch.FloatTensor(self.init_center)
        else:
            init.normal_(self.center, 0, 1)

    def forward(self, X):
        """

        Forward method of Pytorch Module.

        :param torch.tensor X: Pytorch tensor with the size of :math:`[N, D]`, where :math:`N` is the number of samples, :math:`D` is the input dimension.
        :return: Firing level matrix :math:`U` with the size of :math:`[N, R]`.
        """
        frs = self.zr_op(
            -(X.unsqueeze(dim=2) - self.center) ** 2 * (0.5 / (self.sigma ** 2 + self.eps)), dim=1
        )
        frs = functional.softmax(frs, dim=1)
        return frs


class AntecedentShareGMF(Antecedent):
    def __init__(self, in_dim, n_mf=2, high_dim=False, init_center=None, init_sigma=1., eps=1e-8):
        """
        The antecedent part with Gaussian membership function, rules will share the membership
        functions on each feature [2]. The number of rules will be :math:`M^D`, where :math:`M`
        is :code:`n_mf`, :math:`D` is the number of features (:code:`in_dim`).

        :param int in_dim: Number of features :math:`D` of the input.
        :param int n_mf: Number of membership functions :math:`M` of each feature.
        :param bool high_dim: Whether to use the HTSK defuzzification. If :code:`high_dim=True`,
            HTSK is used. Otherwise the original defuzzification is used. More details can be found
            at [1]. TSK model tends to fail on high-dimensional problems, so set :code:`high_dim=True`
            is highly recommended for any-dimensional problems.
        :param numpy.array init_center: Initial center of the Gaussian membership function with
            the size of :math:`[D,M]`.
        :param float init_sigma: Initial :math:`\sigma` of the Gaussian membership function.
        :param float eps: A constant to avoid the division zero error.

        [1] `Cui Y, Wu D, Xu Y. Curse of dimensionality for tsk fuzzy neural networks:
        Explanation and solutions[C]//2021 International Joint Conference on Neural Networks
        (IJCNN). IEEE, 2021: 1-8. <https://arxiv.org/pdf/2102.04271.pdf>`_
        """
        super(AntecedentShareGMF, self).__init__()

        self.in_dim = in_dim
        self.n_mf = n_mf
        self.n_rule = self.n_mf ** self.in_dim
        self.high_dim = high_dim

        self.init_center = check_tensor(init_center, torch.float32) if init_center is not None else None
        self.init_sigma = init_sigma
        self.zr_op = torch.mean if high_dim else torch.sum
        self.eps = eps

        self.__build_model__()

    def __build_model__(self):
        self.center = nn.Parameter(torch.normal(0, 1, size=(self.in_dim, self.n_mf)))
        self.sigma = nn.Parameter(torch.ones(size=(self.in_dim, self.n_mf)) * self.init_sigma)

        self.rule_index = list(itertools.product(*[range(self.n_mf) for _ in range(self.in_dim)]))

        self.reset_parameters()

    def init(self, center, sigma):
        """
        Change the value of :code:`init_center` and :code:`init_sigma`.

        :param numpy.array center: Initial center of the Gaussian membership function
            with the size of :math:`[D,M]`.
        :param float sigma: Initial :math:`\sigma` of the Gaussian membership function.
        """
        center = check_tensor(center, torch.float32)
        self.init_center = center
        self.init_sigma = sigma

        self.reset_parameters()

    def reset_parameters(self):
        """
        Re-initialize all parameters.
        :return:
        """
        init.constant_(self.sigma, self.init_sigma)

        if self.init_center is not None:
            self.center.data[...] = torch.FloatTensor(self.init_center)
        else:
            init.normal_(self.center, 0, 1)

    def forward(self, X):
        """
        Forward method of Pytorch Module.

        :param torch.tensor X: pytorch tensor with the size of :math:`[N, D]`,
            where :math:`N` is the number of samples, :math:`D` is the input dimension.
        :return: Firing level matrix :math:`U` with the size of :math:`[N, R], R=M^D`:.
        """
        zrs = []
        for r in range(self.n_rule):
            mf_index = torch.tensor(self.rule_index[r], device=X.device, dtype=torch.long).unsqueeze(-1)
            center, sigma = torch.gather(self.center, 1, mf_index), torch.gather(self.sigma, 1, mf_index)

            zr = -0.5 * (X.unsqueeze(2) - center) ** 2 / (sigma ** 2 + self.eps)
            zr = self.zr_op(zr, dim=1)
            zrs.append(zr)
        zrs = torch.cat(zrs, dim=1)
        frs = functional.softmax(zrs, dim=1)
        return frs


class AntecedentTriMF(nn.Module):
    """
    Antecedent with triangle membership function
    """
    def __init__(self, in_dim, n_rule, init_center=None, init_left_dist=3., init_right_dist=3., eps=1e-8):
        super(AntecedentTriMF, self).__init__()
        self.in_dim = in_dim
        self.n_rule = n_rule

        self.init_center = check_tensor(init_center, torch.float32) if init_center is not None else None
        self.init_left_dist = math.sqrt(init_left_dist)
        self.init_right_dist = math.sqrt(init_right_dist)
        self.eps = eps

        self.__build_model__()

    def __build_model__(self):
        self.center = nn.Parameter(torch.zeros(size=(self.in_dim, self.n_rule)))
        self.left_dist = nn.Parameter(torch.ones(size=(self.in_dim, self.n_mf)) * self.init_left_dist)
        self.right_dist = nn.Parameter(torch.ones(size=(self.in_dim, self.n_mf)) * self.init_right_dist)

        self.reset_parameters()

    def forward(self, X):
        X = X.unsqueeze(dim=2)
        left_mf = 1 / (self.left_dist ** 2 + self.eps) * X + 1 - self.center / (self.left_dist ** 2 + self.eps)
        right_mf = - 1 / (self.right_dist ** 2 + self.eps) * X + 1 + self.center / (self.right_dist ** 2 + self.eps)
        mf = torch.maximum(
            torch.zeros(1, device=left_mf.device),
            torch.minimum(left_mf, right_mf)
        )
        frs = torch.prod(mf, dim=1)
        frs = frs / (torch.sum(frs, dim=1, keepdim=True) + self.eps)
        return frs

    def init(self, center, left_dist, right_dist):
        """

        Change the value of :code:`init_center` and :code:`init_sigma`.

        :param numpy.array center: Initial center of the Gaussian membership function with the
            size of :math:`[D,R]`. A common way is to run a KMeans clustering and set
            :code:`init_center` as the obtained centers. You can simply run
            :func:`pytsk.gradient_descent.antecedent.antecedent_init_center <antecedent_init_center>`
            to obtain the center.
        :param float left_dist: Initial :math:`\sigma` of the Gaussian membership function.
        :param float right_dist: Initial :math:`\sigma` of the Gaussian membership function.

        """
        center = check_tensor(center, torch.float32)
        self.init_center = center
        self.init_left_dist = math.sqrt(left_dist)
        self.init_right_dist = math.sqrt(right_dist)

        self.reset_parameters()

    def reset_parameters(self):
        """
        Re-initialize all parameters.

        :return:
        """
        init.constant_(self.left_dist, self.init_left_dist ** 2)
        init.constant_(self.right_dist, self.init_right_dist ** 2)

        if self.init_center is not None:
            self.center.data[...] = torch.FloatTensor(self.init_center)
        else:
            init.normal_(self.center, 0, 1)


class AntecedentShareTriMF(nn.Module):
    def __init__(self, in_dim, n_mf=2, init_center=None, init_left_dist=3., init_right_dist=3., eps=1e-8):
        super(AntecedentShareTriMF, self).__init__()

        self.in_dim = in_dim
        self.n_mf = n_mf
        self.n_rule = self.n_mf ** self.in_dim

        self.init_center = check_tensor(init_center, torch.float32) if init_center is not None else None
        self.init_left_dist = math.sqrt(init_left_dist)
        self.init_right_dist = math.sqrt(init_right_dist)
        self.eps = eps

        self.__build_model__()

    def __build_model__(self):
        self.center = nn.Parameter(torch.zeros(size=(self.in_dim, self.n_mf)))
        self.left_dist = nn.Parameter(torch.ones(size=(self.in_dim, self.n_mf)) * self.init_left_dist)
        self.right_dist = nn.Parameter(torch.ones(size=(self.in_dim, self.n_mf)) * self.init_right_dist)

        self.rule_index = list(itertools.product(*[range(self.n_mf) for _ in range(self.in_dim)]))
        self.reset_parameters()

    def reset_parameters(self):
        """
        Re-initialize all parameters.

        :return:
        """
        init.constant_(self.left_dist, self.init_left_dist)
        init.constant_(self.right_dist, self.init_right_dist)

        if self.init_center is not None:
            self.center.data[...] = torch.FloatTensor(self.init_center)
        else:
            init.normal_(self.center, 0, 1)

    def forward(self, X):
        X = X.unsqueeze(dim=2)

        frs = []
        for r in range(self.n_rule):
            mf_index = torch.tensor(self.rule_index[r], device=X.device, dtype=torch.long).unsqueeze(-1)
            c, ld, rd = torch.gather(self.center, 1, mf_index), \
                        torch.gather(self.left_dist, 1, mf_index), \
                        torch.gather(self.right_dist, 1, mf_index)

            left_mf = 1 / (ld ** 2 + self.eps) * X + 1 - c / (ld ** 2 + self.eps)
            right_mf = - 1 / (rd ** 2 + self.eps) * X + 1 + c / (rd ** 2 + self.eps)

            mf = torch.maximum(
                torch.zeros(1, device=left_mf.device),
                torch.minimum(left_mf, right_mf)
            )
            frs.append(torch.prod(mf, dim=1))
        frs = torch.cat(frs, dim=1)
        frs = frs / (torch.sum(frs, dim=1, keepdim=True) + self.eps)
        return frs

